%function D = disparity_population_code(II, n_scales, n_orient, energy_thr, ori_thr, ph_shift, disp1, a)
%
%
% Input:
% II: stereo pairs (grayscale)
% n_scales: number of sclaes (level of pyramid)
% n_oriente : numebro of orientation of the Gabor filters
% ori_thr: threshold on the valid orientation
% ph_shift: phase shifts
% disp1: range of disparities along each orientation
% a: ocular dominance
%
%
%Outputs:
% D: 2D disparity horizontal D(:,:,1) and vertical D(:,:,2)
%
% Authors:
% PSPC-lab - Biophysical and Electronic Engineering Department - University of Genoa
% March 2009 

function D = disparity_population_code(II, n_scales, n_orient, energy_thr, ori_thr, ph_shift, disp1, a)
    [sy1 sx1 st] = size(II);

    if (st~=2)
        error('size(II,3) should equal 2!\n');
    end

    fac = 2^(n_scales-1);

    sy2 = ceil(sy1 ./ fac) .* fac; % target resolution
    sx2 = ceil(sx1 ./ fac) .* fac; % target resolution

    II = [ II; repmat( II(end,:,:), [sy2-sy1 1 1] ) ]; % replicate border row
    II = [ II  repmat( II(:,end,:), [1 sx2-sx1 1] ) ]; % replicate border column

    II = image_pyramid(II,n_scales);

    %%%%%%%%%%%% Level 1 %%%%%%%%%%%%
    [F] = filt_steer_or_gabor( II{1}, 'Gt11B0.0833f0.25.mat');
    G{1,1}=F{1}(:,:,:,1);
    G{1,2}=F{2}(:,:,:,1);
    G(2,:)=shift_in_phase(F{1}(:,:,:,2),F{2}(:,:,:,2),ph_shift);
    clear F
    D = population(G, energy_thr, ori_thr, disp1, a);

    invalid = isnan(D);
    clear E G

    %%%%%%%%%%%% Coarse-to-fine Estimation and Merging %%%%%%%%%%%%%%%%%%%%
    for scale = 2:n_scales,
        D = expand(D.*2);

        [F] = filt_steer_or_gabor(II{scale}, 'Gt11B0.0833f0.25.mat');

        G{1,1} = F{1}(:,:,:,1);
        G{1,2} = F{2}(:,:,:,1);
        G(2,:) = shift_in_phase(F{1}(:,:,:,2), F{2}(:,:,:,2), ph_shift);

        clear F
        G = warp_sequence(G, D);
        Ds = population(G, energy_thr, ori_thr, disp1, a);

        clear G E
        D = merge_disparity(D,Ds);
        invalid = isnan(Ds);
    end % for scale = 2:n_scales,

    D(invalid) = NaN;
    % Remove rows and columns that were added to construct the pyramid
    D(end-(sy2-sy1-1):end,:,:) = [];
    D(:,end-(sx2-sx1-1):end,:) = [];
end % function D = ctf_pop_disparity(II, n_scales, n_orient, energy_thr, ori_thr, file_name, a)
%-------------------------------------------------------------------------------
function [II, x_pix, y_pix] = image_pyramid(II, n_scales)
    [sy sx dummy] = size(II);

    x_pix = cell(1, n_scales);
    y_pix = cell(1, n_scales);

    [ x_pix{n_scales} y_pix{n_scales} ] = meshgrid(1:sx, 1:sy);

    lpf = [1 4 6 4 1]/16;

    tmp = II;
    II = cell(1,n_scales);
    II{n_scales} = tmp;

    for scale = n_scales-1:-1:1
        for frame = 1:2
            tmp(:,:,frame) = conv2b(conv2b(tmp(:,:,frame), lpf, 3), lpf', 3);
        end
        [Ny Nx dummy] = size(tmp);

        tmp = tmp(1:2:Ny,1:2:Nx,:);
        II{scale} = tmp;
        x_pix{scale} = x_pix{scale+1}(1:2:Ny,1:2:Nx);
        y_pix{scale} = y_pix{scale+1}(1:2:Ny,1:2:Nx);
    end
end % function [II, x_pix, y_pix] = image_pyramid(II, n_scales)
%-------------------------------------------------------------------------------
function D = expand(D)
    sy = size(D, 1);
    sx = size(D, 2);

    [X Y] = meshgrid(1:(sx-1)/(2*sx-1):sx, 1:(sy-1)/(2*sy-1):sy);

    D = interp2b(D, X, Y);
end % function D = expand(D)
%-------------------------------------------------------------------------------
function [G_warped] = warp_sequence(G, D)
    n_frame = 2;
    [sy sx n_orient phase_num re_im] = size(G{2});

    G_warped = cell(2, n_frame);
    [X Y] = meshgrid(1:sx, 1:sy);

    D(isnan(D)) = 0;

    % keep first (left) frame
    G_warped{1,1} = G{1,1};
    G_warped{1,2} = G{1,2};

    % warp second frame (right) to first (left)
    Xn = X - D(:,:,1);
    Yn = Y - D(:,:,2);
    G_warped{2,1} = interp2b(double(G{2,1}), Xn, Yn);
    G_warped{2,2} = interp2b(double(G{2,2}), Xn, Yn);
end % function [G_warped] = warp_sequence(G, D)
%-------------------------------------------------------------------------------
function D = merge_disparity(D1,D2)
    invalid1 = isnan(D1);
    invalid2 = isnan(D2);

    D1(invalid1) = 0;
    D2(invalid2) = 0;
    invalid = invalid1 & invalid2;
    % invalid = invalid1; % If unreliable at lower res, do not reattempt at
    % higher res (reliability measure not suited for this). Low res valid
    % estimates are retained even if they are never refined.
    D = D1 + D2;
    D(invalid) = NaN;
end % function D = merge_disparity(D1,D2)
%-------------------------------------------------------------------------------
function D = population(G, energy_thr, ori_thr,disp1, a)
    [sy sx n_orient phase_num re_im] = size(G{2});

    CL_tmp = G{1,1};
    CR_tmp = G{2,1};
    SL_tmp = G{1,2};
    SR_tmp = G{2,2};

    dd  = zeros(sy, sx);
    E   = zeros(sy, sx, n_orient, phase_num);
    Er  = zeros(sy, sx, n_orient, phase_num);
    El  = zeros(sy, sx, n_orient);

    invalid  = zeros(sy, sx, n_orient);
    tmp_real = zeros(sy, sx, n_orient);
    tmp_imag = zeros(sy, sx, n_orient);

    %%ENERGY MODEL
    %%orientation weights%%%
    % w=[1 0.9 0.8 0.65 0 0.65 0.8 0.9]; %uncomment to use this set of weight on
    % orientation. to check (to learn?)
    w = ones(8); %to not consider different weight for different orientation

    %%%ocular dominance
    CL = a*CL_tmp;
    SL = a*SL_tmp;
    CR = (1-a)*CR_tmp;
    SR = (1-a)*SR_tmp;

    for o = 1:n_orient,
        for ph_n = 1:phase_num,
            E(:,:,o,ph_n) = sqrt(  (CL(:,:,o) + CR(:,:,o,ph_n)).^2 ...
                                 + (SL(:,:,o) + SR(:,:,o,ph_n)).^2);
            E(:,:,o,ph_n) = E(:,:,o,ph_n)*w(o);
            El(:,:,o)=(CL(:,:,o).^2 + SL(:,:,o).^2);
            Er(:,:,o,ph_n)=(SR(:,:,o,ph_n).^2 + CR(:,:,o,ph_n).^2);
        end % for ph_n = 1:phase_num,
    end % for o = 1:n_orient,

    % ENERGY THRESHOLD (computed on the monocular energy)
    El = repmat(El,[1 1 1 phase_num]);
    El_norm = El / max(El(:));
    Er_norm = Er / max(Er(:));
    mask = (El_norm > energy_thr) & (Er_norm > energy_thr);

    % ORIENTATION THRESHOLD
    % should not consider the value for which ori_thr values are under threshold
    for i=1:phase_num
        mask3 = mask(:,:,:,i);
        mask4 = (sum(double(mask3),3)>=ori_thr);
        mask5 = repmat(mask4, [1 1 n_orient phase_num]);
        mask6 = mask .* mask5;
    end
    mask = mask6;
    E1 = E.*mask;
    clear mask El Er Er_norm El_norm

    CL = (1-a)*CL_tmp;
    SL = (1-a)*SL_tmp;
    CR = a*CR_tmp;
    SR = a*SR_tmp;
    for o = 1:n_orient,
        for ph_n = 1:phase_num,
            E(:,:,o,ph_n)   = sqrt(   (CL(:,:,o) + CR(:,:,o,ph_n)).^2 ...
                                    + (SL(:,:,o) + SR(:,:,o,ph_n)).^2);
            E(:,:,o,ph_n)   = E(:,:,o,ph_n)*w(o);
            El(:,:,o)       = CL(:,:,o).^2 + SL(:,:,o).^2;
            Er(:,:,o,ph_n)  = SR(:,:,o,ph_n).^2 + CR(:,:,o,ph_n).^2;
        end % for ph_n = 1:phase_num,
    end % for o = 1:n_orient,

    %%ENERGY THRESHOLD (computed on the monocular energy)
    El = repmat(El, [1 1 1 phase_num]);
    El_norm = El / max(El(:));
    Er_norm = Er / max(Er(:));
    mask = (El_norm > energy_thr) & (Er_norm > energy_thr);

    %%ORIENTATION THRESHOLD
    %%should not consider the value for which ori_thr values are under threshold
    for i = 1:phase_num,
        mask3 = mask(:,:,:,i);
        mask4 = (sum(double(mask3), 3) >= ori_thr);
        mask5 = repmat(mask4,[1 1 n_orient phase_num]);
        mask6 = mask.*mask5;
    end % for i = 1:phase_num,

    mask = mask6;
    E2 = E.*mask;
    E = E1 + E2; %just sum the two energy

    E(isnan(E))=0;
    %load(file_name);

    %spatial pooling... comment not to use
    for o = 1:n_orient,
        for i = 1:phase_num,
            %E(:,:,o,i)=medfilt2(E(:,:,o,i),[9 9]);
            E(:,:,o,i) = conv2b(E(:,:,o,i), fspecial('gaussian', [5 5], 1.7), 3);
        end % for i = 1:phase_num,
    end % for o = 1:n_orient,

    %center of gravity for each orientation
    for o = 1:n_orient,
        theta = (o-1)*pi/8;
        for i = 1:sy,
            for j=1:sx,
                E_tmp = E(i,j,o,:);
                E_tmp = E_tmp(:);
                f1 = max(E_tmp);

                if f1==0,
                    dd(i,j) = NaN;
                else
                    a = (E_tmp/f1)';
                    mina = min(a);
                    if ( sum(a)==0 || mina==1 ),
                        dd(i,j) = NaN;
                    else
                        v = (a-mina)/(1-mina);
                        v(v<0.7) = 0;  %threshold on the center of gravity (to learn?)
                        dd(i,j) = sum(disp1.*v)/sum(v);
                    end % if(sum(a)==0 || max(a)-min(a)==0)
                end % if f1==0
            end % for j=1:sx
        end % for i = 1:sy,
        invalid(:,:,o)  = isnan(dd);
        tmp_real(:,:,o) = dd*cos(theta);
        tmp_imag(:,:,o) = dd*sin(theta);
    end % for o = 1:n_orient,

    FV{1}=tmp_real;
    FV{2}=tmp_imag;

    %constraints on orientation
    D = repmat(NaN, [sy sx 2]);
    nc_min = 3;
    for i = 1:sy,
        for j = 1:sx,
            Vx = FV{1}(i,j,:);
            Vy = FV{2}(i,j,:);
            L_2 = Vx.^2+Vy.^2;
            IND2 = isnan(L_2) | L_2<eps;
            L_2(IND2) = [];
            Vx(IND2)  = [];
            Vy(IND2)  = [];
            nc = length(L_2);
            if (nc >= nc_min)
                X = Vx;
                Y = Vy;
                sumX = sum(X);
                sumY = sum(Y);
                sumXYL_2 = sum(X.*Y./L_2);
                sumXXL_2 = sum(X.^2./L_2);
                sumYYL_2 = sum(Y.^2./L_2);
                den = (sumXYL_2^2 - sumXXL_2*sumYYL_2);
                if den==0
                    D(i,j,:) = NaN;
                else % if den==0
                    xr = -(sumX*sumYYL_2-sumY*sumXYL_2) / den;
                    yr = (sumX*sumXYL_2-sumY*sumXXL_2) / den;
                    D(i,j,:) = [xr yr];
                end % if den==0
            end % if (nc >= nc_min)
        end % for j = 1:sx,
    end % for i = 1:sy,
end % function population(G, energy_thr, ori_thr, file_name, a)

%-------------------------------------------------------------------------------
function [F, f0] = filt_steer_or_gabor(I, fname, DC_thr)
    if (nargin<3)
        DC_thr =0; % 1e-10;
    end

    fltr = load(fname);
    switch fltr.type
    case 's'
        F = filt_sep_steer(I, fname, DC_thr);
    case 'g'
        F = filt_sep_gabor(I, fname, DC_thr);
    end
    f0 = fltr.f0;
end % function filt_steer_or_gabor

%-------------------------------------------------------------------------------
function FI_n = filt_sep_gabor(I, fname, DC_thr)
    load(fname, 'taps', 'F');

    n_orient = 8;
    [nr, nc, n_frames] = size(I);
    padsize = floor(taps/2);
    IF = zeros(nr, nc, n_orient, n_frames);

    for frame = 1:n_frames,
        %%%%%%%%%%%% Horizontal and vertical %%%%%%%%%%%%%%
        % 0
        F1Y   = conv2b(I(:,:,frame), F(1,:).', 3);
        even  = conv2b(F1Y, F(2,:), 3);
        odd   = conv2b(F1Y, F(3,:), 3);
        IF(:,:,1,frame) = complex(even, odd);
        clear even odd F1Y;

        % pi/2
        F1X   = conv2b(I(:,:,frame), F(1,:), 3);
        even  = conv2b(F1X, F(2,:).', 3);
        odd   = conv2b(F1X, F(3,:).', 3);
        IF(:,:,5,frame) = complex(even, odd);
        clear even odd F1X;

        %%%%%%%%%%%%% Diagonal %%%%%%%%%%%%%
        F4Y    = conv2b(I(:,:,frame), F(4,:).', 3);
        F4YF4X = conv2b(F4Y, F(4,:), 3);
        F4YF5X = conv2b(F4Y, F(5,:), 3);
        clear F4Y;

        F5Y    = conv2b(I(:,:,frame), F(5,:).', 3);
        F5YF4X = conv2b(F5Y, F(4,:), 3);
        F5YF5X = conv2b(F5Y, F(5,:), 3);
        clear F5Y;

        % pi/4
        even  = F4YF4X - F5YF5X;
        odd   = F5YF4X + F4YF5X;
        IF(:,:,3,frame) = complex(even, odd);

        % 3pi/4
        even  = F4YF4X + F5YF5X;
        odd   = F5YF4X - F4YF5X;
        IF(:,:,7,frame) = complex(even, odd);

        clear even odd F4YF4X F4YF5X F5YF4X F5YF5X;

        %%%%%%%%%%%%%%%% 'In-betweens' %%%%%%%%%%%%%%%%%%
        F8Y    = conv2b(I(:,:,frame), F(8,:).', 3);
        F8YF6X = conv2b(F8Y, F(6,:), 3);
        F8YF7X = conv2b(F8Y, F(7,:), 3);
        clear F8Y;

        F9Y    = conv2b(I(:,:,frame), F(9,:).', 3);
        F9YF6X = conv2b(F9Y, F(6,:), 3);
        F9YF7X = conv2b(F9Y, F(7,:), 3);
        clear F9Y;

        % pi/8
        even  = F8YF6X - F9YF7X;
        odd   = F9YF6X + F8YF7X;
        IF(:,:,2,frame) = complex(even, odd);

        % 7pi/8
        even  = F8YF6X + F9YF7X;
        odd   = F9YF6X - F8YF7X;
        IF(:,:,8,frame) = complex(even, odd);
        clear even odd F8YF6X F8YF7X F9YF6X F9YF7X;

        F6Y    = conv2b(I(:,:,frame), F(6,:).', 3);
        F6YF8X = conv2b(F6Y, F(8,:),3);
        F6YF9X = conv2b(F6Y, F(9,:),3);
        clear F6Y;

        F7Y    = conv2b(I(:,:,frame), F(7,:).', 3);
        F7YF8X = conv2b(F7Y, F(8,:), 3);
        F7YF9X = conv2b(F7Y, F(9,:), 3);
        clear F7Y;

        % 3pi/8
        even  = F6YF8X - F7YF9X;
        odd   = F7YF8X + F6YF9X;
        IF(:,:,4,frame) = complex(even,odd);

        % 5pi/8
        even  = F6YF8X + F7YF9X;
        odd   = F7YF8X - F6YF9X;
        IF(:,:,6,frame) = complex(even,odd);
        clear even odd F6YF8X F6YF9X F7YF8X F7YF9X;
    end % for frame = 1:n_frames,

    invalid = (abs(real(IF))<DC_thr) | (abs(imag(IF))<DC_thr);
    IF(invalid) = NaN;
    FI_n{1} = real(IF);
    FI_n{2} = imag(IF);
end %function filt_sep_gabor(I, fname, DC_thr)

%-------------------------------------------------------------------------------
function F = filt_sep_steer(I, fname, DC_thr)
    load(fname, 'taps', 'G', 'H', 'order');
    n_orient = 8;
    [nr, nc, n_frames] = size(I);
    nG = size(G,1);
    nH = size(H,1);
    F  = zeros(nr, nc, n_orient, n_frames);

    for frame = 1:n_frames,
        % Construct the basic filter responses
        Gb = zeros(nr, nc, nG);
        Hb = zeros(nr, nc, nH);
        for i = 1:nG,
            % X-direction
            f = squeeze(G(i,1,:))';
            T = conv2b(I(:,:,frame),f,3);
            % Y-direction
            f = squeeze(G(i,2,:));
            Gb(:,:,i) = conv2b(T,f,3);
        end

        for i = 1:nH,
            % X-direction
            f = squeeze(H(i,1,:))';
            T = conv2b(I(:,:,frame),f,3);
            % Y-direction
            f = squeeze(H(i,2,:));
            Hb(:,:,i) = conv2b(T,f,3);
        end

        % Steer filters and compute oriented responses
        padsize = floor(taps/2);
        angles = -(0:pi/n_orient:pi-(pi/n_orient)); % negate because gabors
                                                    % rotate in the other direction
        for o = 1:n_orient,
            even = zeros(nr, nc);
            odd  = zeros(nr, nc);
            [Gk,Hk] = gen_steer_coeff(angles(o), order);

            for i = 1:nG
                even = even + Gk(i).*Gb(:,:,i);
            end

            for i = 1:nH
                odd = odd + Hk(i).*Hb(:,:,i);
            end

            invalid = (abs(even)<DC_thr) | (abs(odd)<DC_thr);

            even(invalid) = NaN;
            odd(invalid)  = NaN;
            F(:,:,o,frame) = complex(even, odd);
        end % for o = 1:n_orient
    end % for frame = 1:n_frames
end % function filt_sep_steer

%-------------------------------------------------------------------------------
function [Gk, Hk] = gen_steer_coeff(angle,order)
    % angle is in radians
    if (nargin<2)
        order = 4;
    end
    if ~ismember(order, [2 4])
        error('Order should be equal to 2 or 4.\n');
    end

    c = cos(angle); s = sin(angle);
    switch order
    case 2 % Second order filters
        Gk = [ c.^2; -2.*c.*s;    s.^2 ];
        Hk = [ c.^3; -3.*c.^2.*s; 3.*c.*s.^2; -s.^3];
    case 4 % Fourth order filters
        Gk = [ c.^4; -4.*c.^3.*s; 6.*c.^2.*s.^2;  -4.*c.*s.^3;     s.^4];
        Hk = [ c.^5; -5.*c.^4.*s; 10.*c.^3.*s.^2; -10.*c.^2.*s.^3; 5.*c.*s.^4; -s.^5];
    end
end % function gen_steer_coeff

%-------------------------------------------------------------------------------
function G = shift_in_phase(F_real, F_imag, ph_shift)
      

        G_even_tmp = F_real;%(:,:,orient);
        G_odd_tmp  = F_imag;%(:,:,orient);
        phase = 1;
        for ph = ph_shift,
            G_even(:,:,:,phase) = G_even_tmp*cos(ph) - G_odd_tmp*sin(ph);
            G_odd (:,:,:,phase) = G_odd_tmp *cos(ph) + G_even_tmp*sin(ph);
            phase = phase+1;
        end % for ph = ph_shift,
    G{1,1} = G_even;
    G{1,2} = G_odd;
end % function G=shift_in_phase(F_real,F_imag,file_name)

%-------------------------------------------------------------------------------
function imf = conv2b(im, ker, border)
% function imf = conv2b(im, ker, border);
% CONV2B Two-dimensional convolution with border handling
%   Performs 2-D convolution returning an image with the same
%   dimensions as the input image.
%   The border handling is specified by border:
%       0 -> pad with zeros
%       1 -> periodic convolution
%       2 -> specular borders
% This function is not yet prepared for even sized kernels
    [nky nkx] = size(ker);
    if (~rem(nkx, 2))|(~rem(nky, 2))
        fprintf(1,'\n\t CONV2B WARNING: not specific handling for even sized kernels\n\n');
    end
    % proper treatment for each value of border
    if border==0
        sh = 'same';
    elseif (border==1)|(border==2)|(border==3)
        sh = 'valid';
        Bx = (nkx-1)/2;
        By = (nky-1)/2;
        im = putborde(im, Bx, By, border);
    else
        error('Not a valid value for border')
    end
    % convolving the image
    imf = conv2(im, ker, sh);
end % function imf=conv2b(im, ker, border)

%-------------------------------------------------------------------------------
function imb = putborde(im, Nx, Ny, method);
% function imb = putborde(im, Nx, Ny, method);
% PUTBORDE Puts a border to an image:
%   method=1 -> consider the image periodical
%   method=2 -> specular
%   method=3 -> repeat the edge pixel
    [sy sx] = size(im);
    imb = zeros(sy+2*Ny,sx+2*Nx);
    imb(1+Ny:sy+Ny,1+Nx:sx+Nx) = im;
    switch method,
    case 1,
        imb(1:Ny,1+Nx:sx+Nx)                = im(sy-Ny+1:sy,:);
        imb(sy+Ny+1:sy+2*Ny,1+Nx:sx+Nx)     = im(1:Ny,:);
        imb(1+Ny:sy+Ny,1:Nx)                = im(:,sx-Nx+1:sx);
        imb(1+Ny:sy+Ny,sx+Nx+1:sx+2*Nx)     = im(:,1:Nx);
        imb(1:Ny,1:Nx)                      = im(sy-Ny+1:sy,sx-Nx+1:sx);
        imb(1:Ny,sx+Nx+1:sx+2*Nx)           = im(sy-Ny+1:sy,1:Nx);
        imb(sy+Ny+1:sy+2*Ny,1:Nx)           = im(1:Ny,sx-Nx+1:sx);
        imb(sy+Ny+1:sy+2*Ny,sx+Nx+1:sx+2*Nx)= im(1:Ny,1:Nx);
    case 2,
        imb(Ny:-1:1,1+Nx:sx+Nx)             = im(2:Ny+1,:);
        imb(sy+2*Ny:-1:sy+Ny+1,1+Nx:sx+Nx)  = im(sy-Ny:sy-1,:);
        imb(1+Ny:sy+Ny,Nx:-1:1)             = im(:,2:Nx+1);
        imb(1+Ny:sy+Ny,sx+2*Nx:-1:sx+Nx+1)  = im(:,sx-Nx:sx-1);
        imb(Ny:-1:1,Nx:-1:1)                = im(2:Ny+1,2:Nx+1);
        imb(Ny:-1:1,sx+2*Nx:-1:sx+Nx+1)     = im(2:Ny+1,sx-Nx:sx-1);
        imb(sy+2*Ny:-1:sy+Ny+1,Nx:-1:1)     = im(sy-Ny:sy-1,2:Nx+1);
        imb(sy+2*Ny:-1:sy+Ny+1,sx+2*Nx:-1:sx+Nx+1) = im(sy-Ny:sy-1,sx-Nx:sx-1);
    case 3,
        for k = 1:Nx,
            imb(Ny+1:sy+Ny,k)       = im(:,1);
            imb(Ny+1:sy+Ny,k+sx+Nx) = im(:,sx);
        end
        for k=1:Ny,
            imb(k,Nx+1:sx+Nx)       = im(1,:);
            imb(k+sy+Ny,Nx+1:sx+Nx) = im(sy,:);
        end
    otherwise
        error('Not a valid value for method')
    end % switch method
end % function imb=putborde(im,Nx,Ny,method);

%-------------------------------------------------------------------------------
function output_seq = interp2b(input_seq, col_grid, row_grid)
    dims = size(input_seq);
    rows = dims(1);
    cols = dims(2);
    frames = prod(dims(3:end));

    [new_rows, new_cols] = size(col_grid);

    output_seq = NaN([new_rows new_cols dims(3:end)]);

    % Check for out of range values of col_grid and set to 1
    col_out_ind = find((col_grid<1)|(col_grid>cols));
    if length(col_out_ind)>0,
        col_grid(col_out_ind) = 1;
    end

    % Check for out of range values of row_grid and set to 1
    row_out_ind = find((row_grid<1)|(row_grid>rows));
    if length(row_out_ind)>0,
        row_grid(row_out_ind) = 1;
    end

    % Matrix element indexing
    ndx = floor(row_grid)+floor(col_grid-1)*rows;

    % Compute intepolation parameters, check for boundary value.
    if isempty(col_grid),
        d = col_grid;
    else
        d = find(col_grid==cols);
    end

    col_grid(:) = (col_grid - floor(col_grid));

    if length(d)>0,
        col_grid(d) = col_grid(d)+1;
        ndx(d) = ndx(d)-rows;
    end

    % Compute intepolation parameters, check for boundary value.
    if isempty(row_grid),
        d = row_grid;
    else
        d = find(row_grid==rows);
    end

    row_grid(:) = (row_grid - floor(row_grid));

    if length(d)>0,
        row_grid(d) = row_grid(d)+1;
        ndx(d) = ndx(d)-1;
    end

    % Now interpolate
    onemt = 1 - row_grid;

    for fr_i = 1:frames,
        frame = input_seq(:,:,fr_i);
        output_seq(:,:,fr_i) = ...
            ( frame(ndx).*(onemt) + frame(ndx+1).*row_grid ).*(1-col_grid) + ...
            ( frame(ndx+rows).*(onemt) + frame(ndx+(rows+1)).*row_grid ).*col_grid;
    end % frame-loop
end %function interp2b
